/*
Factory.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import androidx.annotation.Nullable;
import androidx.annotation.NonNull;
import android.content.Context;
import android.os.Build;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.linphone.mediastream.Log;
import org.linphone.mediastream.Version;

/**
* The factory is a singleton object devoted to the creation of all the objects of<br/>
* Liblinphone that cannot be created by {@link Core} itself. <br/>
* <br/>
* It is also used to configure a few behaviors before creating the {@link Core},<br/>
* like the logs verbosity or collection. <br/>
*/
public abstract class Factory {
    static Factory _Factory;

    public static final synchronized Factory instance() {
		try {
			if (_Factory == null) {
				_Factory = new FactoryImpl();
			}
		} catch (Exception e) {
			System.err.println("Cannot instanciate factory");
		}
		return _Factory;
	}

    /**
      * Gets the LoggingService singleton
      */
    abstract public LoggingService getLoggingService();

    abstract public void setDebugMode(boolean enable, String tag);

    abstract public Core getCore(long ptr);

    /**
    * Get the directory where the data resources are located. <br/>
    * <br/>
    * @return The path to the directory where the data resources are located   <br/>
    */
    @NonNull
    abstract public String getDataResourcesDir();

    /**
    * Set the directory where the data resources are located. <br/>
    * <br/>
    * @param path The path where the data resources are located   <br/>
    */
    abstract public void setDataResourcesDir(@NonNull String path);

    /**
    * Returns a bctbx_list_t of all DialPlans. <br/>
    * <br/>
    * @return A list of {@link DialPlan}    <br/>
    */
    @NonNull
    abstract public DialPlan[] getDialPlans();

    /**
    * Get the directory where the image resources are located. <br/>
    * <br/>
    * @return The path to the directory where the image resources are located   <br/>
    */
    @NonNull
    abstract public String getImageResourcesDir();

    /**
    * Set the directory where the image resources are located. <br/>
    * <br/>
    * @param path The path where the image resources are located   <br/>
    */
    abstract public void setImageResourcesDir(@NonNull String path);

    /**
    * Indicates if the storage in database is available. <br/>
    * <br/>
    * @return true if the database storage is available, false otherwise <br/>
    */
    abstract public boolean isDatabaseStorageAvailable();

    /**
    * Indicates if IMDN are available. <br/>
    * <br/>
    * @return true if IDMN are available <br/>
    */
    abstract public boolean isImdnAvailable();

    /**
    * Sets the log collection path. <br/>
    * <br/>
    * @param path the path of the logs   <br/>
    */
    abstract public void setLogCollectionPath(@Nullable String path);

    /**
    * Get the directory where the mediastreamer2 plugins are located. <br/>
    * <br/>
    * @return The path to the directory where the mediastreamer2 plugins are located,<br/>
    * or null if it has not been set.   <br/>
    */
    @Nullable
    abstract public String getMspluginsDir();

    /**
    * Set the directory where the mediastreamer2 plugins are located. <br/>
    * <br/>
    * @param path The path to the directory where the mediastreamer2 plugins are<br/>
    * located   <br/>
    */
    abstract public void setMspluginsDir(@Nullable String path);

    /**
    * Get the directory where the ring resources are located. <br/>
    * <br/>
    * @return The path to the directory where the ring resources are located   <br/>
    */
    @NonNull
    abstract public String getRingResourcesDir();

    /**
    * Set the directory where the ring resources are located. <br/>
    * <br/>
    * @param path The path where the ring resources are located   <br/>
    */
    abstract public void setRingResourcesDir(@NonNull String path);

    /**
    * Get the directory where the sound resources are located. <br/>
    * <br/>
    * @return The path to the directory where the sound resources are located   <br/>
    */
    @NonNull
    abstract public String getSoundResourcesDir();

    /**
    * Set the directory where the sound resources are located. <br/>
    * <br/>
    * @param path The path where the sound resources are located   <br/>
    */
    abstract public void setSoundResourcesDir(@NonNull String path);

    /**
    * Get the list of standard video definitions supported by Linphone. <br/>
    * <br/>
    * @return A list of video definitions.    <br/>
    */
    @NonNull
    abstract public VideoDefinition[] getSupportedVideoDefinitions();

    /**
    * Get the top directory where the resources are located. <br/>
    * <br/>
    * @return The path to the top directory where the resources are located   <br/>
    */
    @NonNull
    abstract public String getTopResourcesDir();

    /**
    * Set the top directory where the resources are located. <br/>
    * <br/>
    * If you only define this top directory, the other resources directory will<br/>
    * automatically be derived form this one. <br/>
    * @param path The path to the top directory where the resources are located   <br/>
    */
    abstract public void setTopResourcesDir(@NonNull String path);

    /**
    * Parse a string holding a SIP URI and create the according {@link Address}<br/>
    * object. <br/>
    * <br/>
    * @param addr A string holding the SIP URI to parse.   <br/>
    * @return A new {@link Address}.   <br/>
    */
    @Nullable
    abstract public Address createAddress(@NonNull String addr);

    /**
    * Creates a {@link AuthInfo} object. <br/>
    * <br/>
    * The object can be created empty, that is with all arguments set to null.<br/>
    * Username, userid, password, realm and domain can be set later using specific<br/>
    * methods. At the end, username and passwd (or ha1) are required. <br/>
    * @param username The username that needs to be authenticated   <br/>
    * @param userid The userid used for authenticating (use null if you don't know<br/>
    * what it is)   <br/>
    * @param passwd The password in clear text   <br/>
    * @param ha1 The ha1-encrypted password if password is not given in clear text.   <br/>
    * @param realm The authentication domain (which can be larger than the sip<br/>
    * domain. Unfortunately many SIP servers don't use this parameter.   <br/>
    * @param domain The SIP domain for which this authentication information is<br/>
    * valid, if it has to be restricted for a single SIP domain.   <br/>
    * @return A {@link AuthInfo} object. linphone_auth_info_destroy must be used to<br/>
    * destroy it when no longer needed. The {@link Core} makes a copy of {@link AuthInfo}<br/>
    * passed through {@link Core#addAuthInfo}.   <br/>
    */
    @NonNull
    abstract public AuthInfo createAuthInfo(@NonNull String username, @Nullable String userid, @Nullable String passwd, @Nullable String ha1, @Nullable String realm, @Nullable String domain);

    /**
    * Creates a {@link AuthInfo} object. <br/>
    * <br/>
    * The object can be created empty, that is with all arguments set to null.<br/>
    * Username, userid, password, realm and domain can be set later using specific<br/>
    * methods. At the end, username and passwd (or ha1) are required. <br/>
    * @param username The username that needs to be authenticated   <br/>
    * @param userid The userid used for authenticating (use null if you don't know<br/>
    * what it is)   <br/>
    * @param passwd The password in clear text   <br/>
    * @param ha1 The ha1-encrypted password if password is not given in clear text.   <br/>
    * @param realm The authentication domain (which can be larger than the sip<br/>
    * domain. Unfortunately many SIP servers don't use this parameter.   <br/>
    * @param domain The SIP domain for which this authentication information is<br/>
    * valid, if it has to be restricted for a single SIP domain.   <br/>
    * @param algorithm The algorithm for encrypting password.   <br/>
    * @return A {@link AuthInfo} object. linphone_auth_info_destroy must be used to<br/>
    * destroy it when no longer needed. The {@link Core} makes a copy of {@link AuthInfo}<br/>
    * passed through {@link Core#addAuthInfo}.   <br/>
    */
    @NonNull
    abstract public AuthInfo createAuthInfo(@NonNull String username, @Nullable String userid, @Nullable String passwd, @Nullable String ha1, @Nullable String realm, @Nullable String domain, @Nullable String algorithm);

    /**
    * Creates an object {@link Buffer}. <br/>
    * <br/>
    * @return a {@link Buffer}   <br/>
    */
    @NonNull
    abstract public Buffer createBuffer();

    /**
    * Creates an object {@link Buffer}. <br/>
    * <br/>
    * @param data the data to set in the buffer   <br/>
    * @param size the size of the data <br/>
    * @return a {@link Buffer}   <br/>
    */
    @NonNull
    abstract public Buffer createBufferFromData(@NonNull byte[] data, int size);

    /**
    * Creates an object {@link Buffer}. <br/>
    * <br/>
    * @param data the data to set in the buffer   <br/>
    * @return a {@link Buffer}   <br/>
    */
    @NonNull
    abstract public Buffer createBufferFromString(@NonNull String data);

    /**
    * Creates an object {@link Config}. <br/>
    * <br/>
    * @param path the path of the config   <br/>
    * @return a {@link Config}   <br/>
    */
    @NonNull
    abstract public Config createConfig(@Nullable String path);

    /**
    * Creates an object {@link Config}. <br/>
    * <br/>
    * @param data the config data   <br/>
    * @return a {@link Config}   <br/>
    */
    @NonNull
    abstract public Config createConfigFromString(@NonNull String data);

    /**
    * Creates an object {@link Config}. <br/>
    * <br/>
    * @param path the path of the config   <br/>
    * @param factoryPath the path of the factory   <br/>
    * @return a {@link Config}   <br/>
    */
    @NonNull
    abstract public Config createConfigWithFactory(@Nullable String path, @Nullable String factoryPath);

    /**
    * Creates an object {@link Content}. <br/>
    * <br/>
    * @return a {@link Content}   <br/>
    */
    @NonNull
    abstract public Content createContent();

    /**
    * Instantiate a {@link Core} object. <br/>
    * <br/>
    * The {@link Core} object is the primary handle for doing all phone actions. It<br/>
    * should be unique within your application. The {@link Core} object is not<br/>
    * started automatically, you need to call {@link Core#start} to that effect. The<br/>
    * returned {@link Core} will be in {@link GlobalState} Ready. Core ressources can<br/>
    * be released using {@link Core#stop} which is strongly encouraged on garbage<br/>
    * collected languages. <br/>
    * @param configPath A path to a config file. If it does not exists it will be<br/>
    * created. The config file is used to store all settings, proxies... so that all<br/>
    * these settings become persistent over the life of the {@link Core} object. It<br/>
    * is allowed to set a null config file. In that case {@link Core} will not store<br/>
    * any settings.   <br/>
    * @param factoryConfigPath A path to a read-only config file that can be used to<br/>
    * store hard-coded preferences such as proxy settings or internal preferences.<br/>
    * The settings in this factory file always override the ones in the normal config<br/>
    * file. It is optional, use null if unneeded.   <br/>
    * @param systemContext A pointer to a system object required by the core to<br/>
    * operate. Currently it is required to pass an android Context on android, pass<br/>
    * null on other platforms.   <br/>
    * @return a {@link Core} object   <br/>
    * see: linphone_core_new_with_config_3() <br/>
    */
    @NonNull
    abstract public Core createCore(@Nullable String configPath, @Nullable String factoryConfigPath, @Nullable Object systemContext);

    /**
    * Instantiate a {@link Core} object with a given LinphoneConfig. <br/>
    * <br/>
    * The {@link Core} object is the primary handle for doing all phone actions. It<br/>
    * should be unique within your application. The {@link Core} object is not<br/>
    * started automatically, you need to call {@link Core#start} to that effect. The<br/>
    * returned {@link Core} will be in {@link GlobalState} Ready. Core ressources can<br/>
    * be released using {@link Core#stop} which is strongly encouraged on garbage<br/>
    * collected languages. <br/>
    * @param config A {@link Config} object holding the configuration for the {@link Core}<br/>
    * to be instantiated.   <br/>
    * @param systemContext A pointer to a system object required by the core to<br/>
    * operate. Currently it is required to pass an android Context on android, pass<br/>
    * null on other platforms.   <br/>
    * @return a {@link Core} object   <br/>
    * see: {@link #createCore} <br/>
    */
    @NonNull
    abstract public Core createCoreWithConfig(@NonNull Config config, @Nullable Object systemContext);

    /**
    * Creates an object LinphoneErrorInfo. <br/>
    * <br/>
    * @return a {@link ErrorInfo} object.   <br/>
    */
    @NonNull
    abstract public ErrorInfo createErrorInfo();

    /**
    * Create a {@link ParticipantDeviceIdentity} object. <br/>
    * <br/>
    * @param address {@link Address} object.   <br/>
    * @param name the name given to the device.   <br/>
    * @return A new {@link ParticipantDeviceIdentity}.   <br/>
    */
    @NonNull
    abstract public ParticipantDeviceIdentity createParticipantDeviceIdentity(@NonNull Address address, @Nullable String name);

    /**
    * Creates an object LinphoneRange. <br/>
    * <br/>
    * @return a {@link Range} object.   <br/>
    */
    @NonNull
    abstract public Range createRange();

    /**
    * Creates an object LinphoneTransports. <br/>
    * <br/>
    * @return a {@link Transports} object.   <br/>
    */
    @NonNull
    abstract public Transports createTransports();

    /**
    * Creates an object {@link TunnelConfig}. <br/>
    * <br/>
    * @return a {@link TunnelConfig}   <br/>
    */
    @NonNull
    abstract public TunnelConfig createTunnelConfig();

    /**
    * Create an empty {@link Vcard}. <br/>
    * <br/>
    * @return a new {@link Vcard}.   <br/>
    */
    @NonNull
    abstract public Vcard createVcard();

    /**
    * Creates an object LinphoneVideoActivationPolicy. <br/>
    * <br/>
    * @return {@link VideoActivationPolicy} object.   <br/>
    */
    @NonNull
    abstract public VideoActivationPolicy createVideoActivationPolicy();

    /**
    * Create a {@link VideoDefinition} from a given width and height. <br/>
    * <br/>
    * @param width The width of the created video definition <br/>
    * @param height The height of the created video definition <br/>
    * @return A new {@link VideoDefinition} object   <br/>
    */
    @NonNull
    abstract public VideoDefinition createVideoDefinition(int width, int height);

    /**
    * Create a {@link VideoDefinition} from a given standard definition name. <br/>
    * <br/>
    * @param name The standard definition name of the video definition to create   <br/>
    * @return A new {@link VideoDefinition} object   <br/>
    */
    @NonNull
    abstract public VideoDefinition createVideoDefinitionFromName(@NonNull String name);

    /**
    * Enables or disables log collection. <br/>
    * <br/>
    * @param state the {@link LogCollectionState} for log collection <br/>
    */
    abstract public void enableLogCollection(LogCollectionState state);

    /**
    * Get the config path. <br/>
    * <br/>
    * @param context used to compute path. Can be null. JavaPlatformHelper on Android<br/>
    * and char *appGroupId on iOS with shared core.   <br/>
    * @return The config path   <br/>
    */
    @NonNull
    abstract public String getConfigDir(@Nullable Object context);

    /**
    * Get the data path. <br/>
    * <br/>
    * @param context used to compute path. Can be null. JavaPlatformHelper on Android<br/>
    * and char *appGroupId on iOS with shared core.   <br/>
    * @return The data path   <br/>
    */
    @NonNull
    abstract public String getDataDir(@Nullable Object context);

    /**
    * Get the download path. <br/>
    * <br/>
    * @param context used to compute path. Can be null. JavaPlatformHelper on Android<br/>
    * and char *appGroupId on iOS with shared core.   <br/>
    * @return The download path   <br/>
    */
    @NonNull
    abstract public String getDownloadDir(@Nullable Object context);

    /**
    * Indicates if the given LinphoneChatRoomBackend is available. <br/>
    * <br/>
    * @param chatroomBackend the {@link ChatRoomBackend} <br/>
    * @return true if the chatroom backend is available, false otherwise <br/>
    */
    abstract public boolean isChatroomBackendAvailable(ChatRoomBackend chatroomBackend);

    /**
    * Select encryption module and set secret material to encrypt the files. <br/>
    * <br/>
    * @param encryptionModule One of the available encryption module for VFS, pick in<br/>
    * the LINPHONE_VFS_ENCRYPTION_* list if set to _UNSET, default bctoolbox VFS is<br/>
    * switch to Standard one <br/>
    * @param secret the secret material used to encrypt the files, can be null for<br/>
    * the _PLAIN module   <br/>
    * @param secretSize size of the secret <br/>
    */
    abstract public void setVfsEncryption(int encryptionModule, @Nullable byte[] secret, int secretSize);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    abstract public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    abstract public Object getUserData();

    /**
      * Gets the native pointer used by this class to make native method calls.
      * @return the nativer pointer, as long
      */
    abstract public long getNativePointer();

    
}

class FactoryImpl extends Factory {

    protected long nativePtr = 0;
    protected transient Object userData = null;

    protected FactoryImpl() {
        nativePtr = init();
    }

    private native long init();

    public long getNativePointer() {
        return nativePtr;
    }

	private static boolean loadOptionalLibrary(String s) {
		try {
			System.loadLibrary(s);
			return true;
		} catch (Throwable e) {
			android.util.Log.w("FactoryImpl", "Unable to load optional library " + s + ": " + e.getMessage());
		}
		return false;
	}

	static {
		System.loadLibrary("c++_shared");
		loadOptionalLibrary("ffmpeg-linphone");
		System.loadLibrary("bctoolbox");
		System.loadLibrary("ortp");
		System.loadLibrary("mediastreamer");
		System.loadLibrary("linphone");
		Version.dumpCapabilities();
	}

	private native Core getCore(long nativePtr, long ptr);
	@Override
	public Core getCore(long ptr) {
	    return getCore(nativePtr, ptr);
	}

    @Override
	public LoggingService getLoggingService() {
		LoggingService l = new LoggingServiceImpl(0);
		return l.get();
	}

    @Override
    public native void setDebugMode(boolean enable, String tag);

    private native String getDataResourcesDir(long nativePtr);
    @Override @NonNull
    synchronized public String getDataResourcesDir()  {
        
        return getDataResourcesDir(nativePtr);
    }

    private native void setDataResourcesDir(long nativePtr, String path);
    @Override
    synchronized public void setDataResourcesDir(@NonNull String path)  {
        
        setDataResourcesDir(nativePtr, path);
    }

    private native DialPlan[] getDialPlans(long nativePtr);
    @Override @NonNull
    synchronized public DialPlan[] getDialPlans()  {
        
        return getDialPlans(nativePtr);
    }

    private native String getImageResourcesDir(long nativePtr);
    @Override @NonNull
    synchronized public String getImageResourcesDir()  {
        
        return getImageResourcesDir(nativePtr);
    }

    private native void setImageResourcesDir(long nativePtr, String path);
    @Override
    synchronized public void setImageResourcesDir(@NonNull String path)  {
        
        setImageResourcesDir(nativePtr, path);
    }

    private native boolean isDatabaseStorageAvailable(long nativePtr);
    @Override
    synchronized public boolean isDatabaseStorageAvailable()  {
        
        return isDatabaseStorageAvailable(nativePtr);
    }

    private native boolean isImdnAvailable(long nativePtr);
    @Override
    synchronized public boolean isImdnAvailable()  {
        
        return isImdnAvailable(nativePtr);
    }

    private native void setLogCollectionPath(long nativePtr, String path);
    @Override
    synchronized public void setLogCollectionPath(@Nullable String path)  {
        
        setLogCollectionPath(nativePtr, path);
    }

    private native String getMspluginsDir(long nativePtr);
    @Override @Nullable
    synchronized public String getMspluginsDir()  {
        
        return getMspluginsDir(nativePtr);
    }

    private native void setMspluginsDir(long nativePtr, String path);
    @Override
    synchronized public void setMspluginsDir(@Nullable String path)  {
        
        setMspluginsDir(nativePtr, path);
    }

    private native String getRingResourcesDir(long nativePtr);
    @Override @NonNull
    synchronized public String getRingResourcesDir()  {
        
        return getRingResourcesDir(nativePtr);
    }

    private native void setRingResourcesDir(long nativePtr, String path);
    @Override
    synchronized public void setRingResourcesDir(@NonNull String path)  {
        
        setRingResourcesDir(nativePtr, path);
    }

    private native String getSoundResourcesDir(long nativePtr);
    @Override @NonNull
    synchronized public String getSoundResourcesDir()  {
        
        return getSoundResourcesDir(nativePtr);
    }

    private native void setSoundResourcesDir(long nativePtr, String path);
    @Override
    synchronized public void setSoundResourcesDir(@NonNull String path)  {
        
        setSoundResourcesDir(nativePtr, path);
    }

    private native VideoDefinition[] getSupportedVideoDefinitions(long nativePtr);
    @Override @NonNull
    synchronized public VideoDefinition[] getSupportedVideoDefinitions()  {
        
        return getSupportedVideoDefinitions(nativePtr);
    }

    private native String getTopResourcesDir(long nativePtr);
    @Override @NonNull
    synchronized public String getTopResourcesDir()  {
        
        return getTopResourcesDir(nativePtr);
    }

    private native void setTopResourcesDir(long nativePtr, String path);
    @Override
    synchronized public void setTopResourcesDir(@NonNull String path)  {
        
        setTopResourcesDir(nativePtr, path);
    }

    private native Address createAddress(long nativePtr, String addr);
    @Override @Nullable
    synchronized public Address createAddress(@NonNull String addr)  {
        
        return (Address)createAddress(nativePtr, addr);
    }

    private native AuthInfo createAuthInfo(long nativePtr, String username, String userid, String passwd, String ha1, String realm, String domain);
    @Override @NonNull
    synchronized public AuthInfo createAuthInfo(@NonNull String username, @Nullable String userid, @Nullable String passwd, @Nullable String ha1, @Nullable String realm, @Nullable String domain)  {
        
        return (AuthInfo)createAuthInfo(nativePtr, username, userid, passwd, ha1, realm, domain);
    }

    private native AuthInfo createAuthInfo2(long nativePtr, String username, String userid, String passwd, String ha1, String realm, String domain, String algorithm);
    @Override @NonNull
    synchronized public AuthInfo createAuthInfo(@NonNull String username, @Nullable String userid, @Nullable String passwd, @Nullable String ha1, @Nullable String realm, @Nullable String domain, @Nullable String algorithm)  {
        
        return (AuthInfo)createAuthInfo2(nativePtr, username, userid, passwd, ha1, realm, domain, algorithm);
    }

    private native Buffer createBuffer(long nativePtr);
    @Override @NonNull
    synchronized public Buffer createBuffer()  {
        
        return (Buffer)createBuffer(nativePtr);
    }

    private native Buffer createBufferFromData(long nativePtr, byte[] data, int size);
    @Override @NonNull
    synchronized public Buffer createBufferFromData(@NonNull byte[] data, int size)  {
        
        return (Buffer)createBufferFromData(nativePtr, data, size);
    }

    private native Buffer createBufferFromString(long nativePtr, String data);
    @Override @NonNull
    synchronized public Buffer createBufferFromString(@NonNull String data)  {
        
        return (Buffer)createBufferFromString(nativePtr, data);
    }

    private native Config createConfig(long nativePtr, String path);
    @Override @NonNull
    synchronized public Config createConfig(@Nullable String path)  {
        
        return (Config)createConfig(nativePtr, path);
    }

    private native Config createConfigFromString(long nativePtr, String data);
    @Override @NonNull
    synchronized public Config createConfigFromString(@NonNull String data)  {
        
        return (Config)createConfigFromString(nativePtr, data);
    }

    private native Config createConfigWithFactory(long nativePtr, String path, String factoryPath);
    @Override @NonNull
    synchronized public Config createConfigWithFactory(@Nullable String path, @Nullable String factoryPath)  {
        
        return (Config)createConfigWithFactory(nativePtr, path, factoryPath);
    }

    private native Content createContent(long nativePtr);
    @Override @NonNull
    synchronized public Content createContent()  {
        
        return (Content)createContent(nativePtr);
    }

    private native Core createCore3(long nativePtr, String configPath, String factoryConfigPath, Object systemContext);
    @Override @NonNull
    synchronized public Core createCore(@Nullable String configPath, @Nullable String factoryConfigPath, @Nullable Object systemContext)  {
        
        return (Core)createCore3(nativePtr, configPath, factoryConfigPath, systemContext);
    }

    private native Core createCoreWithConfig3(long nativePtr, Config config, Object systemContext);
    @Override @NonNull
    synchronized public Core createCoreWithConfig(@NonNull Config config, @Nullable Object systemContext)  {
        
        return (Core)createCoreWithConfig3(nativePtr, config, systemContext);
    }

    private native ErrorInfo createErrorInfo(long nativePtr);
    @Override @NonNull
    synchronized public ErrorInfo createErrorInfo()  {
        
        return (ErrorInfo)createErrorInfo(nativePtr);
    }

    private native ParticipantDeviceIdentity createParticipantDeviceIdentity(long nativePtr, Address address, String name);
    @Override @NonNull
    synchronized public ParticipantDeviceIdentity createParticipantDeviceIdentity(@NonNull Address address, @Nullable String name)  {
        
        return (ParticipantDeviceIdentity)createParticipantDeviceIdentity(nativePtr, address, name);
    }

    private native Range createRange(long nativePtr);
    @Override @NonNull
    synchronized public Range createRange()  {
        
        return (Range)createRange(nativePtr);
    }

    private native Transports createTransports(long nativePtr);
    @Override @NonNull
    synchronized public Transports createTransports()  {
        
        return (Transports)createTransports(nativePtr);
    }

    private native TunnelConfig createTunnelConfig(long nativePtr);
    @Override @NonNull
    synchronized public TunnelConfig createTunnelConfig()  {
        
        return (TunnelConfig)createTunnelConfig(nativePtr);
    }

    private native Vcard createVcard(long nativePtr);
    @Override @NonNull
    synchronized public Vcard createVcard()  {
        
        return (Vcard)createVcard(nativePtr);
    }

    private native VideoActivationPolicy createVideoActivationPolicy(long nativePtr);
    @Override @NonNull
    synchronized public VideoActivationPolicy createVideoActivationPolicy()  {
        
        return (VideoActivationPolicy)createVideoActivationPolicy(nativePtr);
    }

    private native VideoDefinition createVideoDefinition(long nativePtr, int width, int height);
    @Override @NonNull
    synchronized public VideoDefinition createVideoDefinition(int width, int height)  {
        
        return (VideoDefinition)createVideoDefinition(nativePtr, width, height);
    }

    private native VideoDefinition createVideoDefinitionFromName(long nativePtr, String name);
    @Override @NonNull
    synchronized public VideoDefinition createVideoDefinitionFromName(@NonNull String name)  {
        
        return (VideoDefinition)createVideoDefinitionFromName(nativePtr, name);
    }

    private native void enableLogCollection(long nativePtr, int state);
    @Override
    synchronized public void enableLogCollection(LogCollectionState state)  {
        
        enableLogCollection(nativePtr, state.toInt());
    }

    private native String getConfigDir(long nativePtr, Object context);
    @Override @NonNull
    synchronized public String getConfigDir(@Nullable Object context)  {
        
        return getConfigDir(nativePtr, context);
    }

    private native String getDataDir(long nativePtr, Object context);
    @Override @NonNull
    synchronized public String getDataDir(@Nullable Object context)  {
        
        return getDataDir(nativePtr, context);
    }

    private native String getDownloadDir(long nativePtr, Object context);
    @Override @NonNull
    synchronized public String getDownloadDir(@Nullable Object context)  {
        
        return getDownloadDir(nativePtr, context);
    }

    private native boolean isChatroomBackendAvailable(long nativePtr, int chatroomBackend);
    @Override
    synchronized public boolean isChatroomBackendAvailable(ChatRoomBackend chatroomBackend)  {
        
        return isChatroomBackendAvailable(nativePtr, chatroomBackend.toInt());
    }

    private native void setVfsEncryption(long nativePtr, int encryptionModule, byte[] secret, int secretSize);
    @Override
    synchronized public void setVfsEncryption(int encryptionModule, @Nullable byte[] secret, int secretSize)  {
        
        setVfsEncryption(nativePtr, encryptionModule, secret, secretSize);
    }


    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }

    
}
