/*
MagicSearch.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import androidx.annotation.Nullable;
import androidx.annotation.NonNull;

/**
* A {@link MagicSearch} is used to do specifics searchs. <br/>
* <br/>
*/
public interface MagicSearch {
    /**
    * Get the delimiter used for the search. <br/>
    * <br/>
    * @return the delimiter used to find matched filter word   <br/>
    */
    @Nullable
    public String getDelimiter();

    /**
    * Set the delimiter used to find matched filter word. <br/>
    * <br/>
    * @param delimiter delimiter (example "-_.,")   <br/>
    */
    public void setDelimiter(@Nullable String delimiter);

    /**
    * Return whether or not the search is limited. <br/>
    * <br/>
    * @return true if the search is limited, false otherwise <br/>
    */
    public boolean getLimitedSearch();

    /**
    * Enable or disable the limited search. <br/>
    * <br/>
    * @param limited true to limit the search, false otherwise <br/>
    */
    public void setLimitedSearch(boolean limited);

    /**
    * Get the maximum value used to calculate the weight in search. <br/>
    * <br/>
    * @return the maximum value used to calculate the weight in search <br/>
    */
    public int getMaxWeight();

    /**
    * Set the maximum value used to calculate the weight in search. <br/>
    * <br/>
    * @param weight maximum weight <br/>
    */
    public void setMaxWeight(int weight);

    /**
    * Get the minimum value used to calculate the weight in search. <br/>
    * <br/>
    * @return the minimum value used to calculate the weight in search <br/>
    */
    public int getMinWeight();

    /**
    * Set the minimum value used to calculate the weight in search. <br/>
    * <br/>
    * @param weight minimum weight <br/>
    */
    public void setMinWeight(int weight);

    /**
    * Get the number of maximum search result the search will return. <br/>
    * <br/>
    * @return the number of the maximum {@link SearchResult} which will be returned <br/>
    */
    public int getSearchLimit();

    /**
    * Set the number of the maximum SearchResult which will be returned. <br/>
    * <br/>
    * @param limit the maximum number of {@link SearchResult} the search will return <br/>
    */
    public void setSearchLimit(int limit);

    /**
    * Returns whether the delimiter is being used for the search. <br/>
    * <br/>
    * @return if the delimiter search is used <br/>
    */
    public boolean getUseDelimiter();

    /**
    * Enable or disable the delimiter in search. <br/>
    * <br/>
    * @param enable true to use the delimiter, false otherwise <br/>
    */
    public void setUseDelimiter(boolean enable);

    /**
    * Create a sorted list of SearchResult from SipUri, Contact name, Contact<br/>
    * displayname, Contact phone number, which match with a filter word The last item<br/>
    * list will be an address formed with "filter" if a proxy config exist During the<br/>
    * first search, a cache is created and used for the next search Use {@link #resetSearchCache}<br/>
    * to begin a new search. <br/>
    * <br/>
    * @param filter word we search   <br/>
    * @param domain domain which we want to search only  <br/>
    * @return sorted list of      <br/>
    */
    @NonNull
    public SearchResult[] getContactListFromFilter(@Nullable String filter, @Nullable String domain);

    /**
    * Reset the cache to begin a new search. <br/>
    * <br/>
    */
    public void resetSearchCache();

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();

    /**
      * Gets the native pointer used by this class to make native method calls.
      * @return the nativer pointer, as long
      */
    public long getNativePointer();

    public String toString();
}

class MagicSearchImpl implements MagicSearch {

    protected long nativePtr = 0;
    protected transient Object userData = null;

    protected MagicSearchImpl(long ptr) {
        nativePtr = ptr;
    }

    public long getNativePointer() {
        return nativePtr;
    }


    private native String getDelimiter(long nativePtr);
    @Override @Nullable
    synchronized public String getDelimiter()  {
        
        return getDelimiter(nativePtr);
    }

    private native void setDelimiter(long nativePtr, String delimiter);
    @Override
    synchronized public void setDelimiter(@Nullable String delimiter)  {
        
        setDelimiter(nativePtr, delimiter);
    }

    private native boolean getLimitedSearch(long nativePtr);
    @Override
    synchronized public boolean getLimitedSearch()  {
        
        return getLimitedSearch(nativePtr);
    }

    private native void setLimitedSearch(long nativePtr, boolean limited);
    @Override
    synchronized public void setLimitedSearch(boolean limited)  {
        
        setLimitedSearch(nativePtr, limited);
    }

    private native int getMaxWeight(long nativePtr);
    @Override
    synchronized public int getMaxWeight()  {
        
        return getMaxWeight(nativePtr);
    }

    private native void setMaxWeight(long nativePtr, int weight);
    @Override
    synchronized public void setMaxWeight(int weight)  {
        
        setMaxWeight(nativePtr, weight);
    }

    private native int getMinWeight(long nativePtr);
    @Override
    synchronized public int getMinWeight()  {
        
        return getMinWeight(nativePtr);
    }

    private native void setMinWeight(long nativePtr, int weight);
    @Override
    synchronized public void setMinWeight(int weight)  {
        
        setMinWeight(nativePtr, weight);
    }

    private native int getSearchLimit(long nativePtr);
    @Override
    synchronized public int getSearchLimit()  {
        
        return getSearchLimit(nativePtr);
    }

    private native void setSearchLimit(long nativePtr, int limit);
    @Override
    synchronized public void setSearchLimit(int limit)  {
        
        setSearchLimit(nativePtr, limit);
    }

    private native boolean getUseDelimiter(long nativePtr);
    @Override
    synchronized public boolean getUseDelimiter()  {
        
        return getUseDelimiter(nativePtr);
    }

    private native void setUseDelimiter(long nativePtr, boolean enable);
    @Override
    synchronized public void setUseDelimiter(boolean enable)  {
        
        setUseDelimiter(nativePtr, enable);
    }

    private native SearchResult[] getContactListFromFilter(long nativePtr, String filter, String domain);
    @Override @NonNull
    synchronized public SearchResult[] getContactListFromFilter(@Nullable String filter, @Nullable String domain)  {
        
        return getContactListFromFilter(nativePtr, filter, domain);
    }

    private native void resetSearchCache(long nativePtr);
    @Override
    synchronized public void resetSearchCache()  {
        
        resetSearchCache(nativePtr);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }

    @Override
    public String toString() {
        return "Java object [" + super.toString() + "], native pointer [" + String.format("0x%08x", nativePtr) + "]";
    }
}
